

#define _GNU_SOURCE

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>

#ifndef PROGRAM
#   define PROGRAM "mailadmutil"
#endif

#ifndef VERSION
#   define VERSION "0.01"
#endif

#ifndef COPYRIGHT
#   define COPYRIGHT "Copyright 2002 by Jarle (jgaa) Aase"
#endif

#ifndef CONF_FILE
#   define CONF_FILE "/etc/mailadmutil.conf"
#endif

#ifndef DEFAULT_VIRTUSER_ROOT
#   define DEFAULT_VIRTUSER_ROOT "/var/lib/courier/spool/virtusers"
#endif

#ifndef MAKEALIASES_PATH
#   define MAKEALIASES_PATH "/usr/sbin/makealiases"
#endif

#ifndef MAKEHOSTEDDOMAINS_PATH
#   define MAKEHOSTEDDOMAINS_PATH "/usr/sbin/makehosteddomains"
#endif

#ifndef COURIER_PATH
#   define COURIER_PATH "/usr/sbin/courier"
#endif

#ifndef MAKEACCEPTMAILFOR_PATH
#   define MAKEACCEPTMAILFOR_PATH "/usr/sbin/makeacceptmailfor"
#endif

#ifndef MAKESMTPACCESS_APTH
#   define MAKESMTPACCESS_PATH "/usr/sbin/makesmtpaccess"
#endif

/* Global variables */
int debug_flag = 0, verbose_flag = 0,
  mailbox_id = 0, organization_id = 0,
  mbox_uid = 1, mbox_gid = 1, delete_flag = 0;

const char *virtuser_root = DEFAULT_VIRTUSER_ROOT;
const char *makealiases = MAKEALIASES_PATH;
const char *makehosteddomains = MAKEHOSTEDDOMAINS_PATH;
const char *courier_path = COURIER_PATH;
const char *makeacceptmailfor = MAKEACCEPTMAILFOR_PATH;
const char *makesmtpaccess = MAKESMTPACCESS_PATH;
uid_t courier_user = 0;
gid_t courier_group = 0;

static const struct option options[] = {
  {"organization-id", required_argument, NULL, 'o'},
  {"mailbox-id", required_argument, NULL, 'm'},
  {"verbose", no_argument, &verbose_flag, 1},
  {"debug", no_argument, &debug_flag, 1},
  {"version", no_argument, NULL, 'v'},
  {"help", no_argument, NULL, 'h'},
  {NULL, 0, NULL, 0}
};

static const char *usage_str = "Usage:\n"
PROGRAM " [options] command\n"
"Options:\n"
"  -o  --organization-id=#id   Organization id\n"
"  -m  --mailbox-id=#id        Mailbox id\n"
"  -v  --version               Print version and exit\n"
"      --verbose               Be verbose\n"
"      --debug                 Output debugging information\n"
"  -h  --help                  Show this help\n"
"Commands:\n"
"  make-maildir                Create a maildir for a virtual user\n"
"  delete-maildir              Delete a maildir for a virtual user\n"
"  run-makealiases             Run the courier makealiases command as root\n"
"  run-makehosteddomains       Run the courier makehosteddomains command as root\n"
"  run-makeacceptmailfor       Run the courier makeacceptmailfor command as root\n"
"  run-makesmtpaccess          Run the courier makesmtpaccess command as root\n"
"  run-courier-restart         Restart the courier mailserver";

void usage(const char *bad_arg)
{
  if (bad_arg)
    fprintf(stderr, "Bad argument: %s\n", bad_arg);

  printf("%s\n", usage_str);
  exit(-1);
}

int read_conf_file(const char *path)
{
  FILE *fp;
  char buf[1024];
  char token_buf[1024], *token_p;
  const char  *p, *arg_p;
  struct stat st;
  struct passwd *pwd;
  struct group *grp;
  
  if ((fp = fopen(path, "r")) == NULL) {
    if (verbose_flag)
      fprintf(stderr, PROGRAM ": Warning: Unable to "
	      "open the configuration file %s for read!\n(%s)\n"
	      "Using default values in stead.\n",
	      path, strerror(errno));
    return 0;
  }

  if (fstat(fileno(fp), &st) != 0) {
    fprintf(stderr, PROGRAM ": fstat(fp) failed. %s",
	    strerror(errno));
    fclose(fp);
    return -1;
  }

  if ((st.st_uid != 0)
      || (st.st_mode & 0022)) {
    fprintf(stderr, PROGRAM ": The configuration file %s must "
	    "be owned by root and can not be group/world writable!\n",
	    path);
  }

  while((p = fgets(buf, sizeof(buf), fp)) != NULL) {
    while(isspace(*p))
      ++p;
    if (!*p || ('#' == *p))
      continue;
    
    token_p = token_buf;

    while(!isspace(*p)) {
      *token_p++ = *p++; 
    }
    *token_p++ = 0; /* term token buffer */
    arg_p = token_p;

    while(isspace(*p))
      ++p;
    if ('=' == *p)
      ++p;
    while(isspace(*p))
      ++p;
    while(*p && ('\r' != *p) && ('\n' != *p))
      *token_p++ = *p++;
    *token_p = 0; /* term argument buffer */

    if (debug_flag)
      printf(PROGRAM ": (debug) token=%s, argument=%s\n", token_buf, arg_p);

    /* The strdup()'s here introduce an insignificant memory-leak,
       - but the calls are only made once, and the memory are released
       when the program exists, so we really don't care... */
    if (!strcmp(token_buf, "virtuser-root"))
      virtuser_root = strdup(arg_p);
    else if (!strcmp(token_buf, "makealiases-path"))
      makealiases = strdup(arg_p);
    else if (!strcmp(token_buf, "makehosteddomains-path"))
      makehosteddomains = strdup(arg_p);
    else if (!strcmp(token_buf, "makeacceptmailfor-path"))
      makeacceptmailfor = strdup(arg_p);
    else if (!strcmp(token_buf, "makesmtpaccess-path"))
      makesmtpaccess = strdup(arg_p);
    else if (!strcmp(token_buf, "courier-path"))
      courier_path = strdup(arg_p);
    else if (!strcmp(token_buf, "allow-delete")) {
      p = strtok((char *)arg_p, "\t ");
      if (p && !strcmp(p, "yes"))
	delete_flag = 1; /* Allow deletion of dirs */
    }
    else if (!strcmp(token_buf, "mailbox-uid"))
      mbox_uid = atoi(arg_p);
    else if (!strcmp(token_buf, "mailbox-gid"))
      mbox_gid = atoi(arg_p);
    else if (!strcmp(token_buf, "courier-user")) {
      pwd = getpwnam(arg_p);
      if (!pwd) {
	fprintf(stderr, PROGRAM 
		": Error: Failed to lookup user %s from the system passwd file\n",
		arg_p);
	return -1;
      }
      
      courier_user = pwd->pw_uid;
    }
    else if (!strcmp(token_buf, "courier-group")) {
      grp = getgrnam(arg_p);
      if (!grp) {
	fprintf(stderr, PROGRAM 
		": Error: Failed to lookup group %s from the system groups file\n",
		arg_p);
	return -1;
      }

      courier_group = grp->gr_gid;
    }
    else
      fprintf(stderr, PROGRAM ": Warning: Unrecognized token "
	      "in config file %s: %s\n",
	      path, token_buf);
  }

  fclose(fp);
  return 0; /* ok */
}

/* returns: 1 == is dir, 0 == is not dir, -1 == don't exist */
int is_dir(const char *path) {
  struct stat st;

  if (debug_flag)
    printf(PROGRAM ": (debug) evaluating %s... ", path);

  if (stat(path, &st)) {
    if (debug_flag)
      puts("Don't even exist!");
    return -1;
  }
  
  if (S_ISDIR(st.st_mode)) {
    if (debug_flag)
      puts("Is a directory!");
    return 1;
  }

  if (debug_flag)
    puts("Exists, but is not a directory!");
  return 0;
}

int create_dir(const char *path, int mode, uid_t uid, gid_t gid)
{
  if (verbose_flag)
    printf(PROGRAM ": Creating directory %s\n", path);

  if (mkdir(path, mode)) {
    fprintf(stderr, PROGRAM ": Error: Failed to create %s\n%s\n",
	    path, strerror(errno));
    return -1;
  }
  
  if (debug_flag)
    printf(PROGRAM ": (debug) Setting user %d and group %d on path\n",
	   courier_user, courier_group);

  chown(path, uid, gid);
  return 0;
}

int make_maildir()
{
  char path[2048];
  const char *p;
  char *pp = path, *pp_marker;
  struct passwd *pwd;
  struct group *grp;

  if (0 == courier_user) {
    pwd = getpwnam("courier");
    if (!pwd)
      pwd = getpwnam("daemon");
    if (!pwd) {
      fprintf(stderr, PROGRAM 
	      ": Error: Don't know and can't guess the uid of the courier server\n");
      return -1;
    }
    
    courier_user = pwd->pw_uid;
  }

  if (0 == courier_group) {
    grp = getgrnam("courier");
    if (!grp)
      grp = getgrnam("daemon");
    if (!grp) {
      fprintf(stderr, PROGRAM 
	      ": Error: Don't know and can't guess the gid of the courier server\n");
      return -1;
    }
    
    courier_group = grp->gr_gid;
  }

  p = virtuser_root;
  while(*p)
    *pp++ = *p++;

  /* Resolve the root-dir */
  if ((pp > path) && ('/' == pp[-1]))
    --pp;
  *pp = 0;
  
  if (is_dir(path) == -1) {      
    if (create_dir(path, 0755, courier_user, courier_group))
      return -1;
  }

  /* Resolve the organizations root dir */
  pp += sprintf(pp, "/%d", organization_id);

  if (is_dir(path) == -1) {
    if (create_dir(path, 0700, mbox_uid, mbox_gid))
      return -1;
  }

  /* Resolve the users home dir */
  pp += sprintf(pp, "/%d", mailbox_id);

  if (is_dir(path) == -1) {
    if (create_dir(path, 0700, mbox_uid, mbox_gid))
      return -1;
  }  

  /* Resolve the users Maildir */
  pp += sprintf(pp, "/Maildir");

  if (is_dir(path) == -1) {
    if (create_dir(path, 0700, mbox_uid, mbox_gid))
      return -1;
  }

  pp_marker = pp;

  /* Resolve the users Maildir/tmp */
  pp += sprintf(pp, "/tmp");

  if (is_dir(path) == -1) {
    if (create_dir(path, 0700, mbox_uid, mbox_gid))
      return -1;
  }

  /* Resolve the users Maildir/new */
  pp += sprintf(pp_marker, "/new");

  if (is_dir(path) == -1) {
    if (create_dir(path, 0700, mbox_uid, mbox_gid))
      return -1;
  }

  /* Resolve the users Maildir/cur */
  pp += sprintf(pp_marker, "/cur");

  if (is_dir(path) == -1) {
    if (create_dir(path, 0700, mbox_uid, mbox_gid))
      return -1;
  }

  return 0;
}

int delete_maildir()
{
  char buf[2048], *pp;
  const char *p;

  pp = buf;
  p = virtuser_root;
  while(*p)
    *pp++ = *p++;

  /* Resolve the root-dir */
  if ((pp > buf) && ('/' == pp[-1]))
    --pp;
  *pp = 0;

  if (mailbox_id)
    sprintf(pp, "/%d/%d", organization_id, mailbox_id);
  else
    sprintf(pp, "/%d", organization_id);


  if (!delete_flag) {
    printf(PROGRAM ": Ignoring delete request. This feature is disabled in the conf-file.\n"
	 "You must delete the path \"%s\" manually!\n", buf);
    return 0;
  }
  


  if (verbose_flag)
    printf(PROGRAM ": Deleting the path %s\n", buf);
  execl("/bin/rm", "/bin/rm", "-R", buf, NULL); 

  return 0;
}

int run_makealiases()
{
  if (verbose_flag)
    printf(PROGRAM ": Executing %s\n", makealiases);
  execl(makealiases, makealiases, NULL);
  return 0;
}

int run_makehosteddomains()
{
  if (verbose_flag)
    printf(PROGRAM ": Executing %s\n", makehosteddomains);
  execl(makehosteddomains, makehosteddomains, NULL);
  return 0;
}

int run_makeacceptmailfor()
{
  if (verbose_flag)
    printf(PROGRAM ": Executing %s\n", makeacceptmailfor);
  execl(makeacceptmailfor, makeacceptmailfor, NULL);
  return 0;
}

int run_makesmtpaccess()
{
  if (verbose_flag)
    printf(PROGRAM ": Executing %s\n", makesmtpaccess);
  execl(makesmtpaccess, makesmtpaccess, NULL);
  return 0;
}

int run_courier_restart()
{
  if (verbose_flag)
    printf(PROGRAM ": Executing %s restart\n", courier_path);
  execl(courier_path, courier_path, "restart", NULL);
  return 0;
}

int main(int argc, char *argv[])
{
  int c;

  while((c = getopt_long(argc, argv, "hvo:m:",
			 options, NULL)) != -1) {
    switch(c) {
    case 0:
      break;
    case 'v':
      puts(PROGRAM " " VERSION "\n");
      return 0;
    case 'm':
      mailbox_id = atoi(optarg);
      break;
    case 'o':
      organization_id = atoi(optarg);
      break;
    case 'h':
      usage(NULL);
    case ':':
    case '?':
      puts(PROGRAM ": Try --help for help.\n");
      exit(-1);
    default:
      printf("Error: getopt_long() returned unexcpected code %d\n", c);
      return -1;
    }
  }

  if (debug_flag)
    verbose_flag = 1;

  if (optind != (argc -1)) {
    puts(PROGRAM ": Bad number of arguments. Must have one and only one command!\n"
	 "Try " PROGRAM " --help for help\n");
    return -1;
  }

  if (verbose_flag)
    puts(PROGRAM " " VERSION "\n" COPYRIGHT "\n");

  if (read_conf_file(CONF_FILE) != 0)
    return -1;

  if (!strcmp(argv[argc -1], "make-maildir"))
    return make_maildir();

  if (!strcmp(argv[argc -1], "delete-maildir"))
    return delete_maildir();

  if (!strcmp(argv[argc -1], "run-makealiases"))
    return run_makealiases();

  if (!strcmp(argv[argc -1], "run-makehosteddomains"))
    return run_makehosteddomains();

  if (!strcmp(argv[argc -1], "run-makeacceptmailfor"))
    return run_makeacceptmailfor();

  if (!strcmp(argv[argc -1], "run-makesmtpaccess"))
    return run_makesmtpaccess();
 
  if (!strcmp(argv[argc -1], "run-courier-restart"))
    return run_courier_restart();

  fprintf(stderr, PROGRAM ": Unrecognized command %s!\n", 
	  argv[argc-1]);
  return -1;
}
